/*
typedef struct yams_sedge sedge;
typedef struct yams_error error;
typedef struct yams_shashtable shashtable;
typedef struct yams_sstack sstack;
typedef struct yams_sinfo sinfo;
typedef struct yams_soptions soptions;
typedef struct yams_spoint spoint;
typedef struct yams_striangle striangle;
typedef struct yams_squad squad;
typedef struct yams_geomtge geomtge;
typedef struct yams_geomsupp geomsupp;
typedef struct yams_metric metric;
typedef struct yams_smesh smesh;
*/
typedef yams_Edge Edge;
typedef yams_Error Error;
typedef yams_Hashtable Hashtable;
typedef yams_Stack Stack;
typedef yams_Info Info;
typedef yams_Options Options;
typedef yams_Point Point;
typedef yams_Triangle Triangle;
typedef yams_Quad Quad;
typedef yams_Tetra Tetra;
typedef yams_Geomtge Geomtge;
typedef yams_GeomSupp GeomSupp;
typedef yams_Metric Metric;
typedef yams_SurfMesh SurfMesh;

typedef yams_pEdge pEdge;
typedef yams_pHashtable pHashtable;
typedef yams_pStack pStack;
typedef yams_pPoint pPoint;
typedef yams_pTriangle pTriangle;
typedef yams_pQuad pQuad;
typedef yams_pTetra pTetra;
typedef yams_pGeomtge pGeomtge;
typedef yams_pGeomSupp pGeomSupp;
typedef yams_pMetric pMetric;
typedef yams_pSurfMesh pSurfMesh;
